package org.netkernel.module.standard.builtin;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.LinkedHashMap;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.layer0.meta.*;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.INKFBasicContext;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.NKFAsyncRequestHandleImpl;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.endpoint.TransparentOverlayImpl;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;

public class PrivateFilterEndpoint extends TransparentOverlayImpl
{
	public static String KEY_PRIVATE="private";
	
	private static RequestResponseFieldsImpl sJustPrivate=new RequestResponseFieldsImpl(1);
	private static Constructor sEndpointProxyConstructor;
	private static Constructor sRepresentationProxyConstructor;
	private static Constructor sPrototypeProxyConstructor;
	private static Constructor sMetaProxyConstructor;
	
	static
	{	sJustPrivate.put(PrivateFilterEndpoint.KEY_PRIVATE, Boolean.TRUE);

		sEndpointProxyConstructor=createProxy(IEndpointMeta.class);
		sRepresentationProxyConstructor=createProxy(IRepresentationMeta.class);
		sPrototypeProxyConstructor=createProxy(IPrototypeMeta.class);
		sMetaProxyConstructor=createProxy(IMetaRepresentation.class);
	}
	
	private static Constructor createProxy(Class aDerivative)
	{	Class[] classes=new Class[]{aDerivative};
		Class proxyClass = Proxy.getProxyClass(PrivateFilterEndpoint.class.getClassLoader(), classes);
		Constructor result;
		try
		{	result = proxyClass.getConstructor(new Class[] { InvocationHandler.class });
		}
		catch (Exception e)
		{	e.printStackTrace();
			result=null;
		}
		return result;
	}
	
	public PrivateFilterEndpoint(ISpace aInsideSpace) throws NKFException
	{	declareParameters(TransparentOverlayImpl.sPrototypeParameters);
		declareUnhandledExceptionsExpired(false);
		declareThreadSafe();
		setParameterRaw(TransparentOverlayImpl.PARAM_SPACE, aInsideSpace, null);
		declareVerboseUnhandledExceptions(false);
	}
	
	public void onCommissionEndpoint(IKernel aKernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(aKernel, aSpace);
		//we are responsible for commissioning the nested space as we own it now
		getDelegateSpace().onCommissionSpace(aKernel);
	}
	
	/* Override to make non-transparent - we want to filter out some of the parts
	 */
	public ISpaceElements getParts(INKFRequestContext aContext) throws Exception
	{	ISpaceMeta delegateMeta = getDelegateMeta(aContext);
		ISpace delegateSpace= getDelegateSpace();
		IRequestScopeLevel scope = RequestScopeLevelImpl.createOrphanedRootScopeLevel(delegateSpace,aContext.getKernelContext().getRequestScope());
		aContext.getKernelContext().setRequestScope(scope);
	
		//construct space elements, and get all elements meta
		ISpaceElements delegateElements=delegateMeta.getElements();			
		LinkedHashMap<String, IMetaRepresentation> ebid = new LinkedHashMap<String, IMetaRepresentation>();
		TupleList tl=new TupleList(3,delegateElements.size());
		for (int i=0; i<delegateElements.size(); i++)
		{	IIdentifier elementId=delegateElements.getIdentifier(i);
			if (elementId!=null)
			{	try
				{	String elementIdString = elementId.toString();
					IMetaRepresentation meta=aContext.meta(elementIdString);
					boolean isPublic= (meta.getAdditionalFields().getValue(KEY_PRIVATE))==null;
					if (isPublic)
					{	ebid.put(elementIdString, meta);
						tl.put(elementId,delegateSpace,this);
					}
				}
				catch (Exception e)
				{	//aContext.logRaw(ILogger.WARNING, "PrivateFilterEndpoint failure on "+elementId+" for "+getDelegateSpace().toString());
					// safe to ignore
				}
			}
		}
		mElementMetasById=ebid;
		return new SpaceElementsImpl(tl);
	}	
	
	
	public void onResolve(INKFResolutionContext aContext) throws Exception
	{
		boolean debug=aContext.isDebugResolve();
		super.onResolve(aContext);
		if (aContext.isResponseSet())
		{	aContext.delegateResolution(getDelegateSpace());
		}
	}
	
	
	@Override
	public void onRequest(String elementId, INKFRequestContext aContext) throws Exception
	{	Utils.delegateRequestIntoDetectingAsync(getDelegateSpace(), aContext);
	}
	
	/** Create a clone of the Metadata if necessary ensuring that the private field is set
	 * @param aMeta
	 * @return
	 */
	public static IMetaRepresentation cloneAsPrivate(IMetaRepresentation aMeta)
	{
		IRequestResponseFields fields=aMeta.getAdditionalFields();
		Boolean isPrivate=(Boolean)aMeta.getAdditionalFields().getValue(PrivateFilterEndpoint.KEY_PRIVATE);
		if (isPrivate==null) isPrivate=Boolean.FALSE;
		
		IMetaRepresentation result;
		if (isPrivate)
		{	result = aMeta;
		}
		else
		{	RequestResponseFieldsImpl f2;
			if (fields.size()==0)
			{	f2=sJustPrivate;
			}
			else
			{	f2=new RequestResponseFieldsImpl(fields);
				f2.put(PrivateFilterEndpoint.KEY_PRIVATE, Boolean.TRUE);
			}
			
			Constructor constructor;
			if (aMeta instanceof IEndpointMeta)
			{	constructor=sEndpointProxyConstructor;
			}
			else if (aMeta instanceof IRepresentationMeta)
			{	constructor=sRepresentationProxyConstructor;
			}
			else if (aMeta instanceof IPrototypeMeta)
			{	constructor=sPrototypeProxyConstructor;
			}
			else
			{	constructor=sMetaProxyConstructor;
			}
			
			//create a proxy around meta to intercept
			InvocationHandler handler = new CloneAsPrivateHandler(aMeta,f2);
			try
			{	result = (IMetaRepresentation)constructor.newInstance(new Object[]{handler});
			}
			catch (Exception e)
			{	System.out.println(e);
				result=null;
			}
		}
		return result;
	}
	
	private static class CloneAsPrivateHandler implements InvocationHandler
	{
		private final IMetaRepresentation mMeta;
		private final IRequestResponseFields mFields;
		
		public CloneAsPrivateHandler(IMetaRepresentation aMeta, IRequestResponseFields aFields)
		{	mMeta=aMeta;
			mFields=aFields;
		}
		
		public Object invoke(Object proxy, Method method, Object[] args) throws Throwable
		{	if (method.getName().equals("getAdditionalFields"))
			{	return mFields;
			}
			else
			{	return method.invoke(mMeta,args);
			}
		}
	}	
}