package org.netkernel.module.standard.endpoint;

import org.netkernel.container.IKernel;
import org.netkernel.layer0.meta.IPrototypeParameterMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.ISpace;
import org.w3c.dom.Element;


public interface IStandardEndpoint extends IEndpoint
{
	public static final ISpace[] NO_SPACES=new ISpace[0];
	
	/** Gives the endpoint a chance to say it can resolve the request
	 * @param aContext the context containing the request to be resolved
	 * @throws NKFException thrown if there are any problems
	 */
	void onResolve(INKFResolutionContext aContext) throws Exception;
	
	/** Enumerates the parts that this endpoint provides, will only be called after commissioning
	 * @param aContext the context for issuing any necessary subrequests if needed,
	 * this context should accumulate dependencies of any necessary subrequests
	 * @return the parts
	 */
	ISpaceElements getParts(INKFRequestContext aContext) throws Exception;
	
	ISpace[] getSpaces(INKFRequestContext aContext) throws Exception;
	
	/** Sets the value of an initialisation parameter for this endpoint
	 * @param aParameterName
	 * @param aParameterElement 
	 * @param aHook
	 * @throws NKFException thrown if the parameter doesn't exist or the type of the value isn't compatible
	 */
	void setParameter(String aParameterName, Element aParameterElement, IKernel aKernel) throws Exception;
	void setParameterRaw(String aParameterName, Object aValue, IKernel aKernel) throws Exception;
	
	void declareParameters(IPrototypeParameterMeta[] aParameterDefn);
		
	void postCommissionMapping(IRequestScopeLevel aStaticScope) throws Exception;
	void preDecommissionMapping(IRequestScopeLevel aStaticScope) throws Exception;
	
	
}
